//*****************************************************************************
//*****************************************************************************
// OFORTH : MEMORY MANAGEMENT
//*****************************************************************************
//*****************************************************************************

/*
  Copyright (c) 2014, 2018 Franck Bensusan. All rights reserved.

  This file is part of Oforth sources engine.

  Oforth is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation, either version 3
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, see http://www.gnu.org/licenses/.
*/

#include "oforth.h"

#define FIRSTPAGESIZE   2048

extern void sysExceptionFromC(CThread *thread, char *message);
extern int handleMarkStack(CThread *thread, vint nbMarks);
extern void pushMarkStack(CThread *, MyObject *);

extern MYLONG getTick(vint, struct timeb *);

extern Class **allClasses;
extern CThread *mainThread;
extern OforthVM *ovm;

extern Class *Integer, *String;

// Page allocation
extern vint gcTicksByStep, gcAppByStep, gcVerboseLevel;

// Memory size for code
vint codeMem = 0;

//*****************************************************************************


//*****************************************************************************
// COMMON
//*****************************************************************************

//*****************************************************************************
void outOfMemory(CThread *thread)
//*******************************
{
   sysExceptionFromC(thread, "Out of memory error");
}
//*****************************************************************************
// Allocate a chunk of raw persistent memory 
RawPage *allocRawMem(register CThread *thread, register vint size)
//****************************************************************
{
RawPage *page;

    if (size & CELLBOUND) {
        size += sizeof(Cell);
        size >>= CELLSHIFT;
        size <<= CELLSHIFT;
        }

    if (size > 32768) {
      if ((page = (RawPage *) AllocMem(size)) == NULL) outOfMemory(thread);
      thread->heapSize += (size / 1024);
      return page ;
      }

    if (thread->persistentFree < size) {
        if ((thread->persistentPage = (char *) AllocPage) == NULL)
            outOfMemory(thread);
        thread->persistentFree = OPAGESIZE;
        thread->heapSize += KOBYPAGE;
        }
        
    page = (RawPage *) thread->persistentPage;
    thread->persistentPage += size; 
    thread->persistentFree -= size;

    // if (size > 100) dbg("New raw %p of size %d remains %d\n", page, size, thread->persistentFree);
    return page;
}
//*****************************************************************************
// Allocate n cells on the heap. They can be free.
// Raw memory allocated is n+2 cells
Raw *allocCells(CThread *thread, vint n, uchar **last)
//****************************************************
{
register vint size;
register vint index;
Raw *r;

    size = (n << CELLSHIFT) + 2 * sizeof(Cell);
    
    // if (size > 100) dbg("New size %d\n", size);

    if   (n <=   MAXPARAMS)    { index = n; }
    else { if (size <=   256)  { size =   256; index = MAXPARAMS + 1; }
    else { if (size <=   512)  { size =   512; index = MAXPARAMS + 2; }
    else { if (size <=  1024)  { size =  1024; index = MAXPARAMS + 3; }
    else { if (size <=  2048)  { size =  2048; index = MAXPARAMS + 4; }
    else { if (size <=  4096)  { size =  4096; index = MAXPARAMS + 5; }
    else { if (size <=  8192)  { size =  8192; index = MAXPARAMS + 6; }
    else { if (size <= 16384)  { size = 16384; index = MAXPARAMS + 7; }
    else { if (size <= 32768)  { size = 32768; index = MAXPARAMS + 8; }
    else { goto AllocBigCells; } } } } } } } } }
    
    // dbg("Allocating %d cells, size = %d index = %d\n", n, size, index);

    if ((r = thread->raws[index]) != NULL) { thread->raws[index] = r->next; }
    else {
        r = (Raw *) allocRawMem(thread, size);
        r->next = NULL;
        r->objptr = index << MSKSIZESHIFT;
        }

    if (last != NULL) *last = ((uchar *) r) + size;

    return r;

AllocBigCells:
   if (size >= 0x10000000) outOfMemory(thread);
   r = (Raw *) allocRawMem(thread, size);
   r->next   = (Raw *) &ovm;
   r->objptr = size;
   if (last != NULL) { *last = ((uchar *) r) + size; }; 
   return r;
}
//******************************************************************************
// Allocate raw memory 
PTR allocMem(CThread *thread, register vint size, uchar **last)
//*************************************************************
{
Raw *r;

    if (size & CELLBOUND) {
        size += sizeof(Cell);
        size >>= CELLSHIFT;
        size <<= CELLSHIFT;
        }

    r = allocCells(thread, size >> CELLSHIFT, last);
    // dbg("Allocated mem size = %d ==> %p\n", size, r);

    return (PTR) &(r->raw); 
}
//*****************************************************************************
void freeMem(register CThread *thread, register PTR mem)
//******************************************************
{
register Raw *r;
vint heapSize, index;
    
    r = (Raw *) (mem - 2);

    if (r->next == (Raw *) &ovm) {
        heapSize = r->objptr;
        FreeMem(r, heapSize);
        thread->heapSize -= (heapSize / 1024); 
        }
    else { 
        if ((index = r->objptr) & MSKCLASSID) { index &= MSKSIZEALL; }
        index >>= MSKSIZESHIFT;
        r->next = thread->raws[index]; 
        thread->raws[index] = r;
        }
        
   // dbg("Thread %x : free ok for %x\n", thread, mem);
}
//*****************************************************************************
PCell allocDictCell(CThread *thread, vint n)
// *****************************************
{
register vint i;
register PCell dp;

    dp = (PCell) allocRawMem(thread, n * sizeof(Cell));
    for(i = 0; i < n; i++) dp[i] = NULL;
    return dp;
}
//*****************************************************************************
MyObject *allocDictObject(CThread *thread, Class *cl, vint n)
// **********************************************************
{
MyObject *o;

   o = (MyObject *) allocRawMem(thread, n);
   memset(o, 0, n);

   if (cl != NULL) 
        o->objptr = (cl->classPtr | MSKPERSISTENT | MSKALLOCATED ) & MSKMODERUN;
   else
        o->objptr = 0;

   return o;
}
//*****************************************************************************
MyNatural *allocDictInteger(CThread *thread, vint size)
//*****************************************************
{
MyNatural *n;

   n = (MyNatural *) allocDictObject(thread, Integer, sizeof(MyNatural));
   n->buf = (vint *) allocRawMem(thread, size * sizeof(Cell));
   n->free = n->buf + size;
   return n;
}
//*****************************************************************************
MyString *allocDictString(CThread *thread, char *s, vint allocate)
// ***************************************************************
{
MyString *st;
vint len;

   st = (MyString *) allocDictObject(thread, String, sizeof(MyString));
   len = strlen(s);

   if (allocate == OK) {
      st->buf = (uchar *) allocRawMem(thread, len + 1);
      memcpy(st->buf, s, len + 1);
      }
   else { st->buf = (uchar *) s; }

   st->objptr &= (~MSKMUTABLE);
   st->free1 = st->buf + len;
   return st;
}
//*****************************************************************************
MyString *allocPersistentString(CThread *thread, vint size)
// ********************************************************
{
MyString *st;

    st = (MyString *) allocDictObject(thread, String, sizeof(MyString));
    st->buf = (uchar *) allocRawMem(thread, size);
    st->free1 = st->buf;
    *(st->buf) = '\0'; 
    st->alloc = st->buf + size;
    return st;
}
//*****************************************************************************


//*****************************************************************************
// Objects & GC 
//*****************************************************************************

//*****************************************************************************
// Allocate new memory page for objects/raw of size bytes
RawPage *allocPage(CThread *thread, vint size, vint type)
//*******************************************************
{
register RawPage *page;
register uchar *s;
register uchar *o, *omax;
register vint nbobjects;
vint useable;

    // dbg("Allocating page : size = %d, type = %d\n", size, type);

    if (type == FIRST_PAGE) {
        page = allocRawMem(thread, FIRSTPAGESIZE);
        useable = FIRSTPAGESIZE - sizeof(RawPageHeader);
        }
    else {
        if ((page = (RawPage *) AllocPage) == NULL) outOfMemory(thread);
        thread->heapSize += KOBYPAGE;
        useable = OPAGESIZE - sizeof(RawPageHeader);
        }

    page->h.states = page->h.nextstate = s = (uchar *) &(page->body);
    page->h.next   = NULL;
    nbobjects = ((useable / ( size + sizeof(char))) >> 2) << 2; 
    page->h.nbObj = page->h.nbFree = nbobjects; 
    page->h.objects = o = page->h.states + nbobjects;

    // Setting objects state address
    omax = o + size * nbobjects;
    for(; o < omax; o += size) { *((PCell) o) = (Cell) s; *s = GCFREE; s++; }

    //dbg("Thread %x :Page %x, type %d, size=%d, nb=%d states = %x, objects = %x\n",
        //thread, page, type, size,page->h.nbObj, page->h.states, page->h.objects);

    return page;
}
//*****************************************************************************
void GCSweepNextObjects(CThread *thread, register uchar level, sint nb)
//*********************************************************************
{
register vint newFree;
register uchar state, *states, *endState;
vint index;
RawPage *rp, *rpfirst, *rpprev;
MyObject *o;

   // dbg("Thread %x : sweep attributes, index = %d\n", TOCINT(thread->id), 
      // thread->gcSweepIndexObject);
   for(index = thread->gcSweepIndexObject; index <= MAXATTRIBUTE; index++) {
      if ((rpfirst = thread->all_obj[index]) == NULL) continue;
      thread->cur_obj[index] = rpfirst;
      rp = thread->gcSweepPage; 
      thread->gcSweepPage = NULL;

      if (rp == NULL) { rp = rpprev = rpfirst; }
      else { rpprev = thread->gcSweepPrevPage; }
      
      while(rp != NULL) {
        // If page is empty
        if (rp->h.nbObj == rp->h.nbFree) {
            // If first page, always keep it
            if (rp == rpfirst) { rpprev = rp; rp = rp->h.next; continue; }

            // If just after first page and empty, free it
            if (rp == rpfirst->h.next) {
               rpfirst->h.next = rp->h.next;
               FreeMem(rp, OPAGESIZE); 
               thread->heapSize -= KOBYPAGE;
               // dbg("Freeing obj index %d page %x\n", index, rp);
               rp = rpfirst->h.next; 
               nb -= 10; continue;
               }

            // Otherwise, if empty, move page just after the first page
            rpprev->h.next = rp->h.next;
            rp->h.next = rpfirst->h.next;
            rpfirst->h.next = rp;
            thread->gcFreeIndex[index] += rp->h.nbFree;
            rp = rpprev->h.next;
            nb--; continue;
            }

         states = rp->h.nextstate = rp->h.states;
         newFree = 0;

         for(endState = states + rp->h.nbObj; states < endState; states++) {
            state = *states; 
            if (state == GCFREE || state >= level) continue;
            if (state & (GCUSED | GCNEW)) { *states &= ~(GCUSED | GCNEW); continue; }

            *states = GCFREE;
            newFree++;

            if (state & GCFINALIZE) {
               o = (MyObject *)((char *) rp->h.objects 
                     + (((states - rp->h.states) * index) << CELLSHIFT)
                     + sizeof(Cell));
               (allClasses[o->objptr & MSKCLASSID]->finalFunction)(thread, o);
               nb -= 5;
               }
            }

         rp->h.nbFree += newFree;
         thread->gcFreeIndex[index] += rp->h.nbFree;
         nb -= (rp->h.nbObj >> 2);
         rpprev = rp; rp = rp->h.next;

         if (nb <= 0) {
            // dbg("Thread %x : stop attributes %d\n", TOCINT(thread->id), index);
            if (rp == NULL) index++;
            thread->gcSweepIndexObject = index;
            thread->gcSweepPage = rp;
            thread->gcSweepPrevPage = rpprev;
            return;
            }
         }
      }
   
   // dbg("Thread %x : End of attributes %d\n", TOCINT(thread->id), index);
   thread->gcSweepIndexObject = index;
   thread->gcSweepPage = NULL;
}
//*****************************************************************************
void checkImmutable(register CThread *thread, MyObject *o)
//********************************************************
{
vint objp;

   if (!ISINTNIL(o)) {
      objp = o->objptr;
      if (objp & MSKMUTABLE) sysExceptionFromC(thread, "Immutable rule violation");
      }
}
//*****************************************************************************
void setPersistent(register MyObject *o)
//**************************************
{
register vint objp;
register MyObject **param, **endparam;
register PTR mem, memmax;
uchar *pstate;

   // dbg("o = %x\n", o);
   if (ISINTNIL(o)) return;
   // dbg("Setting not removeable for %x %x : %s\n", o, o->objptr, objectClass(o)->name->buf);
   objp = o->objptr;
   if (objp & MSKPERSISTENT) return;

   o->objptr |= MSKPERSISTENT;

   if (!(objp & MSKALLOCATED)) {
       pstate = (uchar *) *(((uchar **) o) - 1);
       *pstate |= GCPERSISTENT;
       }

   param    = (MyObject **) &(((PCell) o)[(objp & MSKSIZEHIDDEN) >> MSKHSIZE]);
   endparam = (MyObject **) &(((PCell) o)[(objp & MSKSIZEALL)    >> MSKSIZE]);
   // dbg("Param = %x, endparam = %x\n", param, endparam);
   while(param < endparam) { param++; setPersistent(*param); }

   if (objp & MSKARRAYGC) {
      mem = (PTR) ((MyArray *) o)->buf;
      if (mem != nil) {
         for(memmax = ((MyArray *) o)->free; mem < memmax; mem++) 
			   setPersistent((MyObject *) *mem);
         }
      }
   // dbg("End setting flag %d for %x\n", flag, o);
}
//*****************************************************************************
void GCSaveTaskRoots(CThread *thread, MyTask *t)
//**********************************************
{
register PCell first, last;
register vint nbVar, max;
PCell vt;

    // Store Task runnable object on mark stack
    pushMarkStack(thread, t->runnable);

    // Store task variables valueson mark stack
    max = (vint) t->vtvar[0]; vt = t->vtvar;
    for (nbVar = 1; nbVar <= max; nbVar++)
        pushMarkStack(thread, (MyObject *) vt[nbVar]); 

    // Store data stack objects on mark stack
    last = t->dsbase;
    for(first = t->ds; first<last;first++) pushMarkStack(thread,(MyObject *) *first);

    // Store return stack objects on mark stack
    if (t->rs != NULL) {
        first = t->rs + 1;
        last  = t->rsbase;

        while(first <= last) {
            if ((vint) *first == CFRAMEUP) 
                { first += ((((vint) *(first + 1)) >> 2) + 3); continue; }

            nbVar = ((vint) *first) / sizeof(Cell) - 1; 
            first++;
            if (nbVar > 1024) { continue; }
            while(nbVar > 0) { 
               pushMarkStack(thread, (MyObject *) *first); 
               first++; nbVar--;
               }
            first++;
            }
         }
}
//******************************************************************************
void GCSaveChannelRoots(CThread *thread, MyChannel *ch)
//*****************************************************
{
register PTR mem;
MyTask *t;
vint i, index, size, max;

    // Save channel on return stack
    // pushMarkStack(thread, (MyObject *) ch);

    // Save tasks waiting for event on the channel the channel
    for(t = ch->waitNotEmpty; t != NULL; t = t->waitNext)
        GCSaveTaskRoots(thread, t);

    for(t = ch->waitNotFull; t != NULL; t = t->waitNext)
        GCSaveTaskRoots(thread, t);

    // Save objects into the channel
    if ((mem = ch->buf) == nil) return;

    index = ch->first; 
    size  = ch->size; 
    max   = ch->maxSize;

    for(i = 0; i < size; i++) {
        pushMarkStack(thread, (MyObject *) mem[index]);
        index++; 
        if (index == max) index = 0;
        }

}
//*****************************************************************************
void GCSaveThreadRoots(CThread *thread, CThread *th)
//**************************************************
{
MyTask *task;
MyChannel *ch;

   // Save resumable tasks roots
   for(task = th->resumableTasks; task != NULL; task = task->resumableNext)
      GCSaveTaskRoots(thread, task);

   // Save sleeping tasks
   for(task = th->sleepingTasks; task != NULL; task = task->waitNext)
      GCSaveTaskRoots(thread, task);

   // Save channels roots
   for(ch = th->channels; ch != NULL; ch = ch->next)
      GCSaveChannelRoots(thread, ch);
}
//*****************************************************************************
CThread *GCThreadToSnapshot(CThread *thread, OforthVM *vm)
//*******************************************************
{
CThread *th;

   LockRes(vm)
      if (vm->initSnapshot == KO) {
         vm->initSnapshot = OK;
         vm->currentThreadGC = mainThread;
         }

      th = vm->currentThreadGC;
      if (th != NULL) vm->currentThreadGC = th->next;
   RelRes(vm);

   return th;
}
//*****************************************************************************
void GCLog(CThread *thread, char *log, int taskAfter)
//***************************************************
{
vint t ;

    t = (vint) (getTick(OK, NULL) - thread->vm->tick);
    
    if (taskAfter == OK) { printf("%s:%d(%d);TASK; ", log, thread->id, t); }
    else { printf("%s:%d(%d); ", log, thread->id, t); }

}
//*****************************************************************************
void GCHandleRequest(CThread *thread)
//***********************************
{
int markDone;
CThread *toMark;
OforthVM *vm;
vint i, nb;

   vm = thread->vm;

   // if GC is asked for this thread, initialize and handle snapshot phase
   if (thread->state & ASK_GC) {

      thread->state &= ANSWERED_GC;
      thread->gcTicks = gcAppByStep;
      thread->gcSweepIndexObject = 1;
      thread->gcSweepPage = NULL;
      for(i = 1; i <= MAXATTRIBUTE; i++) thread->gcFreeIndex[i] = 0;

      if (gcVerboseLevel > 1) GCLog(thread, "GCW_BEGIN", KO);

      if (vm->state != VM_SNAPSHOT) {
          thread->gcPhase = GC_SWEEPING;
          if (gcVerboseLevel > 1) GCLog(thread, "GCW_BYPASS", KO);
          goto LSweep;
          }

      thread->gcPhase = GC_SNAPSHOT;

      LockDec(vm->nbWorkersToSynchronize);
      LockInc(vm->nbWorkersSnapshot);
      LockInc(vm->nbWorkersMarking);
      LockInc(vm->nbWorkersEndMarking);

      Barrier;


      // Wait until all workers are synchronized to begin snapshot
#ifdef __MTHREAD_ENABLE__
      if (vm->nbWorkersToSynchronize != 0) {
         if (gcVerboseLevel > 1) { GCLog(thread, "SNAPW_WAIT", KO); }

         while(vm->nbWorkersToSynchronize != 0) { 
            THREAD_YIELD;
            }
         }
#endif

      if (gcVerboseLevel > 1) GCLog(thread, "SNAPW_BEGIN", KO);

      // Snapshot threads, including sleeping threads
      while ((toMark = GCThreadToSnapshot(thread, vm)) != NULL) 
          GCSaveThreadRoots(thread, toMark);

      // Wait until snapshot is finished
      nb = LockDec(vm->nbWorkersSnapshot);
      if (gcVerboseLevel > 1) GCLog(thread, "SNAPW_END", KO);

      // When last worker has ended snapshot, vm jump du MARK phase
      if (nb == 1) {
         vm->state = VM_MARK; 
         if (gcVerboseLevel > 0) GCLog(thread, "MARK_BEGIN", KO); 
         }

#ifdef __MTHREAD_ENABLE__
      while(vm->state == VM_SNAPSHOT) { THREAD_YIELD; }
#endif

      thread->gcPhase = GC_MARKING;
      thread->gcTicks = 0;

      if (gcVerboseLevel > 1) GCLog(thread, "MARKW_BEGIN", OK);
      return;
      }

   // MARK STEPS
   if (thread->gcPhase == GC_MARKING) {
      if (thread->gcTicks > 0) return;
      thread->gcTicks = gcAppByStep;

      if (gcVerboseLevel > 2) GCLog(thread, "MARKW_STEP", KO);

      // Mark step : if not finished, return to run tasks
      if (thread->state == GLOBAL_GC)
           markDone = handleMarkStack(thread, gcTicksByStep * 2);
      else markDone = handleMarkStack(thread, gcTicksByStep);

      if (markDone != OK) { 
         if (gcVerboseLevel > 2) GCLog(thread, "MARKW_STEPEND", OK);
         return;
         }

      if (gcVerboseLevel > 1) GCLog(thread, "MARKW_END", KO);

      LockDec(vm->nbWorkersMarking);
      // dbg("MARKDONE %d\n", TOCINT(thread->id));

#ifdef __MTHREAD_ENABLE__
      while(vm->state == VM_SNAPSHOT) { THREAD_YIELD; }
#endif

      thread->gcPhase = GC_MARKDONE; 
      }

   // SYNCHRONIZE AT THE END OF MARK PHASE
   if (thread->gcPhase == GC_MARKDONE) {

      // While mark is not complete, check Mark Stack and return to tasks
      if (vm->nbWorkersMarking != 0) 
         { handleMarkStack(thread, gcTicksByStep); return; }

      // If mark is done for all threads, synchronize all amrking workers before
      // sweep phase
      while (handleMarkStack(thread, gcTicksByStep) == KO);

      thread->gcPhase = GC_SWEEPING;
      thread->gcTicks = 0;
      nb = LockDec(vm->nbWorkersEndMarking);

      if (nb == 1) { 
         vm->state = VM_SWEEP; 
         if (gcVerboseLevel > 1) { GCLog(thread, "MARK_END", KO); }
         }

      LockInc(vm->nbWorkersToSynchronize);
      }

   // SWEEP STEPS
LSweep:
   if (thread->gcPhase == GC_SWEEPING) {

#ifdef __MTHREAD_ENABLE__
      if (vm->state != VM_SWEEP) {
         if (gcVerboseLevel > 1) { GCLog(thread, "SWEEPW_WAIT", KO); }
         while(vm->state != VM_SWEEP) { THREAD_YIELD; }
         }
#endif
      
      if (thread->gcTicks > 0) return;
      thread->gcTicks = gcAppByStep;

      if (gcVerboseLevel > 2) GCLog(thread, "SWEEPW_STEP", KO);

      // Sweep objects
      if (thread->gcSweepIndexObject <= MAXATTRIBUTE) {
         if (thread->state == GLOBAL_GC) 
              GCSweepNextObjects(thread, GCPERSISTENT, gcTicksByStep * 7); 
         else GCSweepNextObjects(thread, GCUSEDIMMUT , gcTicksByStep * 7); 

         if (gcVerboseLevel > 2) GCLog(thread, "SWEEPW_END", OK);
         return;
         }

      // Sweep is done. Worker returns to RUN state
      if (gcVerboseLevel > 1) GCLog(thread, "SWEEPW_END", OK);

      thread->state = THREAD_RUNNING;
      thread->gcPhase = GC_NOGC;
      thread->gcTicks = YIELD_TICKS;
      nb = LockDec(vm->nbWorkersSweeping);

      if (gcVerboseLevel > 1) GCLog(thread, "GCW_END", OK);

      // If this is the last worker to sweep, GC is done : returns VM to RUN state
      if (nb == 1) { 
         vm->state = VM_RUN; 
         if (gcVerboseLevel > 0) { GCLog(thread, "GC_END", KO); printf("\n"); }
         }
      }
}
//*****************************************************************************
void GC(int type, TICK tick)
//**************************
{
CThread *th;
vint i;

   ovm->tick = tick;
   ovm->initSnapshot = KO;
   ovm->nbWorkersMarking = ovm->nbWorkersEndMarking = 0;
   ovm->nbWorkersSweeping = ovm->nbWorkers;
   ovm->state = VM_SNAPSHOT;
   Barrier;
   
   if (gcVerboseLevel > 0) {
      if (type == GLOBAL_GC) { i = 'G'; } else { i = 'R'; }
#ifdef __OWIN__
         printf("\nGC_EVENT %c(%d):%I64d\n", i, ovm->nbWorkers, tick);
#endif

#ifdef __OLNX__
         printf("\nGC_EVENT %c(%d):%lld\n", i, ovm->nbWorkers, tick);
#endif
      }

   for(th = mainThread; th != NULL; th = th->next) th->state = type + ASK_GC;
}
//*****************************************************************************

